import asyncio
import binascii
from dataclasses import dataclass, field
from decimal import Decimal
from random import randrange
from typing import Any, Dict, Final, List, Optional, cast

from pydantic import BaseModel, ConfigDict, SecretStr, field_validator
from pydantic.v1 import Field
from xrpl.asyncio.account import get_next_valid_seq_number
from xrpl.asyncio.clients import Client, XRPLRequestFailureException
from xrpl.asyncio.transaction import XRPLReliableSubmissionException
from xrpl.asyncio.transaction.main import (
    _LEDGER_OFFSET,
    _calculate_fee_per_transaction_type,
    _get_network_id_and_build_version,
    _tx_needs_networkID,
)
from xrpl.models import Request, Response, Transaction, TransactionMetadata, Tx
from xrpl.models.requests.request import LookupByLedgerRequest, RequestMethod
from xrpl.models.utils import require_kwargs_on_init
from xrpl.utils.txn_parser.utils import NormalizedNode, normalize_nodes
from xrpl.utils.txn_parser.utils.order_book_parser import (
    _get_change_amount,
    _get_fields,
    _get_offer_status,
    _get_quality,
    _group_offer_changes_by_account,
)
from xrpl.utils.txn_parser.utils.types import AccountOfferChange, AccountOfferChanges, OfferChange
from yaml.representer import SafeRepresenter

from hummingbot.client.config.config_data_types import BaseConnectorConfigMap, ClientFieldData
from hummingbot.client.config.config_validators import validate_with_regex
from hummingbot.connector.exchange.xrpl import xrpl_constants as CONSTANTS
from hummingbot.core.data_type.trade_fee import TradeFeeSchema

CENTRALIZED = True
EXAMPLE_PAIR = "XRP-USD"

DEFAULT_FEES = TradeFeeSchema(
    maker_percent_fee_decimal=Decimal("0"),
    taker_percent_fee_decimal=Decimal("0"),
    buy_percent_fee_deducted_from_returns=True,
)
_REQ_ID_MAX: Final[int] = 1_000_000


def get_order_book_changes(metadata: TransactionMetadata) -> List[AccountOfferChanges]:
    """
    Parse all order book changes from a transaction's metadata.

    Args:
        metadata: Transactions metadata.

    Returns:
        All offer changes caused by the transaction.
        The offer changes are grouped by their owner accounts.
    """
    return compute_order_book_changes(metadata)


def _get_offer_change(node: NormalizedNode) -> Optional[AccountOfferChange]:
    status = _get_offer_status(node)
    taker_gets = _get_change_amount(node, "TakerGets")
    taker_pays = _get_change_amount(node, "TakerPays")
    account = _get_fields(node, "Account")
    sequence = _get_fields(node, "Sequence")
    flags = _get_fields(node, "Flags")
    # if required fields are None: return None
    if (
        taker_gets is None
        or taker_pays is None
        or account is None
        or sequence is None
        # or flags is None # flags can be None
    ):
        return None

    expiration_time = _get_fields(node, "Expiration")
    quality = _get_quality(taker_gets, taker_pays)
    offer_change = OfferChange(
        flags=flags,
        taker_gets=taker_gets,
        taker_pays=taker_pays,
        sequence=sequence,
        status=status,
        maker_exchange_rate=quality,
    )
    if expiration_time is not None:
        offer_change["expiration_time"] = expiration_time
    return AccountOfferChange(maker_account=account, offer_change=offer_change)


def compute_order_book_changes(
    metadata: TransactionMetadata,
) -> List[AccountOfferChanges]:
    """
    Compute the offer changes from offer objects affected by the transaction.

    Args:
        metadata: Transactions metadata.

    Returns:
        All offer changes caused by the transaction.
        The offer changes are grouped by their owner accounts.
    """
    normalized_nodes = normalize_nodes(metadata)
    offer_nodes = [node for node in normalized_nodes if node["LedgerEntryType"] == "Offer"]
    offer_changes = []
    for node in offer_nodes:
        change = _get_offer_change(node)
        if change is not None:
            offer_changes.append(change)
    return _group_offer_changes_by_account(offer_changes)


def convert_string_to_hex(s, padding: bool = True):
    if len(s) > 3:
        hex_str = binascii.hexlify(s.encode()).decode()
        if padding:
            while len(hex_str) < 40:
                hex_str += "00"  # pad with zeros to reach 160 bits (40 hex characters)
        return hex_str.upper()

    return s


def get_token_from_changes(token_changes: [Dict[str, Any]], token: str) -> Optional[Dict[str, Any]]:
    for token_change in token_changes:
        if token_change.get("currency") == token:
            return token_change
    return None


class XRPLMarket(BaseModel):
    base: str
    quote: str
    base_issuer: str
    quote_issuer: str
    trading_pair_symbol: Optional[str] = None

    def __repr__(self):
        return str(self.model_dump())

    def get_token_symbol(self, code: str, issuer: str) -> Optional[str]:
        if self.trading_pair_symbol is None:
            return None

        if code.upper() == self.base.upper() and issuer.upper() == self.base_issuer.upper():
            return self.trading_pair_symbol.split("-")[0]

        if code.upper() == self.quote.upper() and issuer.upper() == self.quote_issuer.upper():
            return self.trading_pair_symbol.split("-")[1]

        return None


def represent_xrpl_market(dumper, data):
    return dumper.represent_dict(data.dict())


SafeRepresenter.add_representer(XRPLMarket, represent_xrpl_market)


@require_kwargs_on_init
@dataclass(frozen=True)
class Ledger(Request, LookupByLedgerRequest):
    """
    Retrieve information about the public ledger.
    `See ledger <https://xrpl.org/ledger.html>`_
    """

    method: RequestMethod = field(default=RequestMethod.LEDGER, init=False)
    transactions: bool = False
    expand: bool = False
    owner_funds: bool = False
    binary: bool = False
    queue: bool = False


async def autofill(
    transaction: Transaction, client: Client, signers_count: Optional[int] = None, try_count: int = 0
) -> Transaction:
    """
    Autofills fields in a transaction. This will set `sequence`, `fee`, and
    `last_ledger_sequence` according to the current state of the server this Client is
    connected to. It also converts all X-Addresses to classic addresses.

    Args:
        transaction: the transaction to be signed.
        client: a network client.
        signers_count: the expected number of signers for this transaction.
            Only used for multisigned transactions.

    Returns:
        The autofilled transaction.
    """
    try:
        transaction_json = transaction.to_dict()
        if not client.network_id:
            await _get_network_id_and_build_version(client)
        if "network_id" not in transaction_json and _tx_needs_networkID(client):
            transaction_json["network_id"] = client.network_id
        if "sequence" not in transaction_json:
            sequence = await get_next_valid_seq_number(transaction_json["account"], client)
            transaction_json["sequence"] = sequence
        if "fee" not in transaction_json:
            fee = int(await _calculate_fee_per_transaction_type(transaction, client, signers_count))
            fee = fee * CONSTANTS.FEE_MULTIPLIER
            transaction_json["fee"] = str(fee)
        if "last_ledger_sequence" not in transaction_json:
            ledger_sequence = await get_latest_validated_ledger_sequence(client)
            transaction_json["last_ledger_sequence"] = ledger_sequence + _LEDGER_OFFSET
        if "source_tag" not in transaction_json:
            transaction_json["source_tag"] = CONSTANTS.HBOT_SOURCE_TAG_ID
        return Transaction.from_dict(transaction_json)
    except Exception as e:
        if try_count < CONSTANTS.VERIFY_TRANSACTION_MAX_RETRY:
            return await autofill(transaction, client, signers_count, try_count + 1)
        else:
            raise Exception(f"Autofill failed: {e}")


async def get_latest_validated_ledger_sequence(client: Client) -> int:
    """
    Returns the sequence number of the latest validated ledger.

    Args:
        client: The network client to use to send the request.

    Returns:
        The sequence number of the latest validated ledger.

    Raises:
        XRPLRequestFailureException: if the rippled API call fails.
    """

    request = Ledger(ledger_index="validated")
    request_dict = request.to_dict()
    request_dict["id"] = f"{request.method}_{randrange(_REQ_ID_MAX)}"
    request_with_id = Ledger.from_dict(request_dict)

    response = await client._request_impl(request_with_id)
    if response.is_successful():
        return cast(int, response.result["ledger_index"])

    raise XRPLRequestFailureException(response.result)


_LEDGER_CLOSE_TIME: Final[int] = 1


async def _sleep(seconds: int):
    await asyncio.sleep(seconds)


async def _wait_for_final_transaction_outcome(
    transaction_hash: str, client: Client, prelim_result: str, last_ledger_sequence: int
) -> Response:
    """
    The core logic of reliable submission.  Polls the ledger until the result of the
    transaction can be considered final, meaning it has either been included in a
    validated ledger, or the transaction's LastLedgerSequence has been surpassed by the
    latest ledger sequence (meaning it will never be included in a validated ledger).
    """
    await _sleep(_LEDGER_CLOSE_TIME)

    current_ledger_sequence = await get_latest_validated_ledger_sequence(client)

    if current_ledger_sequence >= last_ledger_sequence:
        raise XRPLReliableSubmissionException(
            f"The latest validated ledger sequence {current_ledger_sequence} is "
            f"greater than LastLedgerSequence {last_ledger_sequence} in "
            f"the transaction. Prelim result: {prelim_result}"
        )

    # query transaction by hash
    transaction_response = await client._request_impl(Tx(transaction=transaction_hash))
    if not transaction_response.is_successful():
        if transaction_response.result["error"] == "txnNotFound":
            """
            For the case if a submitted transaction is still
            in queue and not processed on the ledger yet.
            """
            return await _wait_for_final_transaction_outcome(
                transaction_hash, client, prelim_result, last_ledger_sequence
            )
        else:
            raise XRPLRequestFailureException(transaction_response.result)

    result = transaction_response.result
    if "validated" in result and result["validated"]:
        # result is in a validated ledger, outcome is final
        return_code = result["meta"]["TransactionResult"]
        if return_code != "tesSUCCESS":
            raise XRPLReliableSubmissionException(f"Transaction failed: {return_code}")
        return transaction_response

    # outcome is not yet final
    return await _wait_for_final_transaction_outcome(transaction_hash, client, prelim_result, last_ledger_sequence)


class XRPLConfigMap(BaseConnectorConfigMap):
    connector: str = Field(default="xrpl", const=True, client_data=None)
    xrpl_secret_key: SecretStr = Field(
        default=...,
        client_data=ClientFieldData(
            prompt=lambda cm: "Enter your XRPL wallet secret key",
            is_secure=True,
            is_connect_key=True,
            prompt_on_new=True,
        ),
    )

    wss_node_url: str = Field(
        default="wss://xrplcluster.com/",
        client_data=ClientFieldData(
            prompt=lambda cm: "Enter your XRPL Websocket Node URL",
            is_secure=False,
            is_connect_key=True,
            prompt_on_new=True,
        ),
    )

    wss_second_node_url: str = Field(
        default="wss://s1.ripple.com/",
        client_data=ClientFieldData(
            prompt=lambda cm: "Enter your second XRPL Websocket Node URL",
            is_secure=False,
            is_connect_key=True,
            prompt_on_new=True,
        ),
    )

    wss_third_node_url: str = Field(
        default="wss://s2.ripple.com/",
        client_data=ClientFieldData(
            prompt=lambda cm: "Enter your third XRPL Websocket Node URL",
            is_secure=False,
            is_connect_key=True,
            prompt_on_new=True,
        ),
    )

    custom_markets: Dict[str, XRPLMarket] = Field(
        default={
            "SOLO-XRP": XRPLMarket(
                base="SOLO",
                quote="XRP",
                base_issuer="rsoLo2S1kiGeCcn6hCUXVrCpGMWLrRrLZz",
                quote_issuer="",
            )
        },
        client_data=ClientFieldData(
            prompt=lambda mi: "Enter custom markets: ", is_connect_key=True, prompt_on_new=False
        ),
    )
    model_config = ConfigDict(title="xrpl")

    @field_validator("xrpl_secret_key", mode="before")
    @classmethod
    def validate_xrpl_secret_key(cls, v: str):
        pattern = r"^s[A-HJ-NP-Za-km-z1-9]*$"
        error_message = "Invalid XRPL wallet secret key. Secret key should be a base 58 string and start with 's'."
        ret = validate_with_regex(v, pattern, error_message)
        if ret is not None:
            raise ValueError(ret)
        return v

    @field_validator("wss_node_url", mode="before")
    @classmethod
    def validate_wss_node_url(cls, v: str):
        pattern = r"^(wss://)[\w.-]+(:\d+)?(/[\w.-]*)*$"
        error_message = "Invalid node url. Node url should be in websocket format."
        ret = validate_with_regex(v, pattern, error_message)
        if ret is not None:
            raise ValueError(ret)
        return v

    @field_validator("wss_second_node_url", mode="before")
    @classmethod
    def validate_wss_second_node_url(cls, v: str):
        pattern = r"^(wss://)[\w.-]+(:\d+)?(/[\w.-]*)*$"
        error_message = "Invalid node url. Node url should be in websocket format."
        ret = validate_with_regex(v, pattern, error_message)
        if ret is not None:
            raise ValueError(ret)
        return v

    @field_validator("wss_third_node_url", mode="before")
    @classmethod
    def validate_wss_third_node_url(cls, v: str):
        pattern = r"^(wss://)[\w.-]+(:\d+)?(/[\w.-]*)*$"
        error_message = "Invalid node url. Node url should be in websocket format."
        ret = validate_with_regex(v, pattern, error_message)
        if ret is not None:
            raise ValueError(ret)
        return v


KEYS = XRPLConfigMap.construct()



from enum import Enum
from typing import Tuple

from bidict import bidict

from hummingbot.core.api_throttler.data_types import LinkedLimitWeightPair, RateLimit
from hummingbot.core.data_type.in_flight_order import OrderState

EXCHANGE_NAME = "Coinbase Advanced Trade"

CHANGELOG_URL = "https://docs.cdp.coinbase.com/advanced-trade/docs/changelog"
LATEST_UPDATE = "2024-SEP-04"
# curl https://docs.cdp.coinbase.com/advanced-trade/docs/changelog | md5sum
CHANGELOG_HASH = "4825e9a0e67b58f6be38f7e411637b87"

COINBASE_ADVANCED_TRADE_CLASS_PREFIX = "CoinbaseAdvancedTrade"

DEFAULT_DOMAIN = "com"

HBOT_ORDER_ID_PREFIX = "CBAT-"
MAX_ORDER_ID_LEN = 32
HBOT_BROKER_ID = "Hummingbot"

# Base URL
BASE_URL = "api.coinbase.com"
SIGNIN_URL = "https://api.coinbase.{domain}/v2"
REST_URL = "https://api.coinbase.{domain}/api/v3"
WSS_URL = "wss://advanced-trade-ws.coinbase.{domain}"
USER_WSS_URL = "wss://advanced-trade-ws-user.coinbase.{domain}"

# Coinbase SDK version: 1.7.0 - The official SDK requires websocket >=12
USER_AGENT = "coinbase-advanced-py/1.7.0"

# Coinbase Signin API endpoints
EXCHANGE_RATES_USD_EP = "/exchange-rates"
EXCHANGE_RATES_QUOTE_EP = "/exchange-rates?currency={quote_token}"
EXCHANGE_RATES_QUOTE_LIMIT_ID = "ExchangeRatesQuote"
CURRENCIES_EP = "/currencies"
CRYPTO_CURRENCIES_EP = "/currencies/crypto"

SIGNIN_ENDPOINTS = {
    EXCHANGE_RATES_USD_EP,
    EXCHANGE_RATES_QUOTE_LIMIT_ID,
    CURRENCIES_EP,
    CRYPTO_CURRENCIES_EP,
}

# Private API endpoints
SERVER_TIME_EP = "/brokerage/time"
ALL_PAIRS_EP = "/brokerage/products"
PAIR_TICKER_EP = "/brokerage/products/{product_id}"
PAIR_TICKER_RATE_LIMIT_ID = "PairTicker"
PAIR_TICKER_24HR_EP = "/brokerage/products/{product_id}/ticker"
PAIR_TICKER_24HR_RATE_LIMIT_ID = "PairTicker24Hr"
ORDER_EP = "/brokerage/orders"
BATCH_CANCEL_EP = "/brokerage/orders/batch_cancel"
GET_ORDER_STATUS_EP = "/brokerage/orders/historical/{order_id}"
GET_ORDER_STATUS_RATE_LIMIT_ID = "GetOrderStatus"
GET_STATUS_BATCH_EP = "/brokerage/orders/historical/batch"
FILLS_EP = "/brokerage/orders/historical/fills"
TRANSACTIONS_SUMMARY_EP = "/brokerage/transaction_summary"
ACCOUNTS_LIST_EP = "/brokerage/accounts"
ACCOUNT_EP = "/brokerage/accounts/{account_uuid}"
ACCOUNT_RATE_LIMIT_ID = "Account"
SNAPSHOT_EP = "/brokerage/product_book"

PRIVATE_REST_ENDPOINTS = {
    ALL_PAIRS_EP,
    PAIR_TICKER_RATE_LIMIT_ID,
    PAIR_TICKER_24HR_RATE_LIMIT_ID,
    ORDER_EP,
    BATCH_CANCEL_EP,
    GET_ORDER_STATUS_RATE_LIMIT_ID,
    GET_STATUS_BATCH_EP,
    FILLS_EP,
    TRANSACTIONS_SUMMARY_EP,
    ACCOUNTS_LIST_EP,
    ACCOUNT_RATE_LIMIT_ID,
    SNAPSHOT_EP,
}

PUBLIC_REST_ENDPOINTS = {
    SERVER_TIME_EP,
}

WS_HEARTBEAT_TIME_INTERVAL = 30


class WebsocketAction(Enum):
    SUBSCRIBE = "subscribe"
    UNSUBSCRIBE = "unsubscribe"


# https://docs.cdp.coinbase.com/advanced-trade/docs/ws-channels
# TODO: this is not exclusively ORDER SUBSCRIPTION, please review the naming
WS_ORDER_SUBSCRIPTION_KEYS: Tuple[str, ...] = ("level2", "market_trades")
WS_ORDER_SUBSCRIPTION_CHANNELS: bidict[str, str] = bidict({"l2_data": "order_book_diff", "market_trades": "trade"})
WS_MAX_MSG_SIZE = 8 * 1024 * 1024

WS_USER_SUBSCRIPTION_KEYS: str = "user"
# WS_USER_SUBSCRIPTION_KEYS: Tuple[str, ...] = ("user",)
WS_USER_SUBSCRIPTION_CHANNELS: bidict[str, str] = bidict({k: k for k in WS_USER_SUBSCRIPTION_KEYS})

WS_OTHERS_SUBSCRIPTION_KEYS: Tuple[str, ...] = ("ticker", "ticker_batch", "status", "candles")
WS_OTHERS_SUBSCRIPTION_CHANNELS: bidict[str, str] = bidict({k: k for k in WS_OTHERS_SUBSCRIPTION_KEYS})

# CoinbaseAdvancedTrade params
SIDE_BUY = "BUY"
SIDE_SELL = "SELL"

# Rate Limit Type
PRIVATE_REST_REQUESTS = "PRIVATE_REST_REQUESTS"
MAX_PRIVATE_REST_REQUESTS_S = 30

PUBLIC_REST_REQUESTS = "PUBLIC_REST_REQUESTS"
MAX_PUBLIC_REST_REQUESTS_S = 10

SIGNIN_REQUESTS = "SIGNIN_REQUESTS"
MAX_SIGNIN_REQUESTS_H = 10000

WSS_REQUESTS = "WSS_REQUESTS"
MAX_WSS_REQUESTS_S = 750

# Rate Limit time intervals
ONE_SECOND = 1
ONE_MINUTE = 60
ONE_HOUR = 3600
ONE_DAY = 86400

# Order States
ORDER_STATE = {
    "OPEN": OrderState.OPEN,
    "PENDING": OrderState.PENDING_CREATE,
    "FILLED": OrderState.FILLED,
    "CANCELLED": OrderState.CANCELED,
    "EXPIRED": OrderState.FAILED,
    "FAILED": OrderState.FAILED,
    # Not directly from exchange
    "PARTIALLY_FILLED": OrderState.PARTIALLY_FILLED,
}
# Oddly, order can be in unknown state ???
ORDER_STATUS_NOT_FOUND_ERROR_CODE = "UNKNOWN_ORDER_STATUS"

PRIVATE_REST_RATE_LIMITS = [
    RateLimit(limit_id=endpoint,
              limit=MAX_PRIVATE_REST_REQUESTS_S,
              time_interval=ONE_SECOND,
              linked_limits=[LinkedLimitWeightPair(PRIVATE_REST_REQUESTS, 1)]) for endpoint in PRIVATE_REST_ENDPOINTS]

PUBLIC_REST_RATE_LIMITS = [
    RateLimit(limit_id=endpoint,
              limit=MAX_PUBLIC_REST_REQUESTS_S,
              time_interval=ONE_SECOND,
              linked_limits=[LinkedLimitWeightPair(PRIVATE_REST_REQUESTS, 1)]) for endpoint in PUBLIC_REST_ENDPOINTS]

SIGNIN_RATE_LIMITS = [
    RateLimit(limit_id=endpoint,
              limit=MAX_SIGNIN_REQUESTS_H,
              time_interval=ONE_HOUR,
              linked_limits=[LinkedLimitWeightPair(SIGNIN_REQUESTS, 1)]) for endpoint in SIGNIN_ENDPOINTS]

RATE_LIMITS = [
    RateLimit(limit_id=PRIVATE_REST_REQUESTS, limit=MAX_PRIVATE_REST_REQUESTS_S, time_interval=ONE_SECOND),
    RateLimit(limit_id=PUBLIC_REST_REQUESTS, limit=MAX_PUBLIC_REST_REQUESTS_S, time_interval=ONE_SECOND),
    RateLimit(limit_id=SIGNIN_REQUESTS, limit=MAX_SIGNIN_REQUESTS_H, time_interval=ONE_HOUR),
    RateLimit(limit_id=WSS_REQUESTS, limit=MAX_WSS_REQUESTS_S, time_interval=ONE_SECOND),
]

RATE_LIMITS.extend(PRIVATE_REST_RATE_LIMITS)
RATE_LIMITS.extend(PUBLIC_REST_RATE_LIMITS)
RATE_LIMITS.extend(SIGNIN_RATE_LIMITS)




from decimal import Decimal

from pydantic.v1 import Field

from hummingbot.client.config.config_data_types import ClientFieldData
from hummingbot.client.config.strategy_config_data_types import BaseTradingStrategyMakerTakerConfigMap


class CrossExchangeMiningConfigMap(BaseTradingStrategyMakerTakerConfigMap):
    strategy: str = Field(default="cross_exchange_mining", client_data=None)

    min_profitability: Decimal = Field(
        default=...,
        description="The minimum estimated profitability required to open a position.",
        ge=-100.0,
        le=100.0,
        client_data=ClientFieldData(
            prompt=lambda mi: "What is the minimum profitability for you to make a trade? (Enter 1 to indicate 1%)",
            prompt_on_new=True,
        ),
    )
    order_amount: Decimal = Field(
        default=...,
        description="The amount of base currency for the strategy to maintain over exchanges.",
        ge=0.0,
        client_data=ClientFieldData(
            prompt=lambda mi: CrossExchangeMiningConfigMap.order_amount_prompt(mi),
            prompt_on_new=True,
        )
    )

    balance_adjustment_duration: float = Field(
        default=Decimal("5"),
        description="Time interval to rebalance portfolio >>> ",
        client_data=ClientFieldData(
            prompt=lambda mi: "Time interval between subsequent portfolio rebalances ",
            prompt_on_new=True,
        ),
    )

    slippage_buffer: Decimal = Field(
        default=Decimal("5.0"),
        description="Allowed slippage to fill ensure taker orders are filled.",
        ge=0.0,
        le=100.0,
        client_data=ClientFieldData(
            prompt=lambda mi: (
                "How much buffer do you want to add to the price to account for slippage for taker orders "
                "Enter 1 to indicate 1%"
            ),
            prompt_on_new=True,
        ),
    )

    min_prof_tol_low: Decimal = Field(
        default=Decimal("0.05"),
        description="Tolerance below min prof to cancel order.",
        ge=0.0,
        le=100.0,
        client_data=ClientFieldData(
            prompt=lambda mi: (
                "What percentage below the min profitability do you want to cancel the set order"
                "Enter 0.1 to indicate 0.1%"
            ),
            prompt_on_new=True,
        ),
    )

    min_prof_tol_high: Decimal = Field(
        default=Decimal("0.05"),
        description="Tolerance above min prof to cancel order.",
        ge=0.0,
        le=100.0,
        client_data=ClientFieldData(
            prompt=lambda mi: (
                "What percentage above the min profitability level do you want to cancel the set order"
                "Enter 0.1 to indicate 0.1%"
            ),
            prompt_on_new=True,
        ),
    )
    volatility_buffer_size: int = Field(
        default=Decimal("120"),
        description="The period in seconds to calulate volatility over: ",
        client_data=ClientFieldData(
            prompt=lambda mi: "The period in seconds to calulate volatility over: ",
            prompt_on_new=True,
        ),
    )

    min_prof_adj_timer: float = Field(
        default=Decimal("3600"),
        description="Time interval to adjust min profitability over",
        client_data=ClientFieldData(
            prompt=lambda mi: "Time interval to adjust min profitability over by using results of previous trades in last 24 hrs",
            prompt_on_new=True,
        ),
    )
    min_order_amount: Decimal = Field(
        default=Decimal("0.0"),
        description="What is the minimum order amount required for bid or ask orders?: ",
        ge=0.0,
        client_data=ClientFieldData(
            prompt=lambda mi: (
                    "What is the minimum order amount required for bid or ask orders?: "
            ),
            prompt_on_new=True,
        ),
    )
    rate_curve: Decimal = Field(
        default=Decimal("1.0"),
        description="Multiplier for rate curve for the adjustment of min profitability based on previous trades over last 24 hrs: ",
        ge=0.0,
        client_data=ClientFieldData(
            prompt=lambda mi: (
                    "Multiplier for rate curve for the adjustment of min profitability based on previous trades over last 24 hrs: "
            ),
            prompt_on_new=True,
        ),
    )
    trade_fee: Decimal = Field(
        default=Decimal("0.25"),
        description="Complete trade fee covering both taker and maker trades: ",
        ge=0.0,
        client_data=ClientFieldData(
            prompt=lambda mi: (
                    "Complete trade fee covering both taker and maker trades: "
            ),
            prompt_on_new=True,
        ),
    )
    # === prompts ===

    @classmethod
    def order_amount_prompt(cls, model_instance: 'CrossExchangeMiningConfigMap') -> str:
        trading_pair = model_instance.maker_market_trading_pair
        base_asset, quote_asset = trading_pair.split("-")
        return f"The amount of {base_asset} for the strategy to maintain in wallet over exchanges (Will autobalance by buying or selling to maintain amount).?"



import logging
from decimal import Decimal
from typing import Any, Dict, List, Tuple, Union

import pandas as pd

from hummingbot.client.settings import AllConnectorSettings
from hummingbot.connector.derivative.position import Position
from hummingbot.core.clock import Clock
from hummingbot.core.data_type.common import OrderType, PositionAction, PositionMode, PositionSide, TradeType
from hummingbot.core.data_type.limit_order import LimitOrder
from hummingbot.core.data_type.order_candidate import OrderCandidate, PerpetualOrderCandidate
from hummingbot.core.network_iterator import NetworkStatus
from hummingbot.logger import HummingbotLogger
from hummingbot.strategy.hedge.hedge_config_map_pydantic import HedgeConfigMap
from hummingbot.strategy.market_trading_pair_tuple import MarketTradingPairTuple
from hummingbot.strategy.strategy_py_base import StrategyPyBase
from hummingbot.strategy.utils import order_age

hedge_logger = None


class HedgeStrategy(StrategyPyBase):
    """
    This strategy contains 2 mode of hedging.
    1. Hedge by amount
    2. Hedge by value

    1. Hedge by amount
    The strategy will hedge by amount by calculating the amount to hedge by each asset.
    The amount of asset to hedge is calculated by the following formula:
    for each asset in the hedge market pair,
        amount_to_hedge = sum of asset amount with the same base asset * hedge_ratio + hedge asset amount
    The amount of asset to hedge must be greater than the minimum trade size to be traded.

    2. Hedge by value
    The strategy will hedge by value by calculating the amount of asset to hedge.
    The amount of asset to hedge is calculated by the following formula:
    amount_to_hedge = sum of asset value of all market pairs * hedge_ratio + hedge asset value
    The amount of asset to hedge must be greater than the minimum trade size to be traded.
    """

    @classmethod
    def logger(cls) -> HummingbotLogger:
        global hedge_logger
        if hedge_logger is None:
            hedge_logger = logging.getLogger(__name__)
        return hedge_logger

    def __init__(
        self,
        config_map: HedgeConfigMap,
        hedge_market_pairs: List[MarketTradingPairTuple],
        market_pairs: List[MarketTradingPairTuple],
        offsets: Dict[MarketTradingPairTuple, Decimal],
        status_report_interval: float = 900,
        max_order_age: float = 5,
        enable_auto_set_position_mode: bool = True,
    ):
        """
        Initializes the hedge strategy.
        :param hedge_market_pair: Market pair to hedge.
        :param market_pairs: Market pairs to trade.
        :param hedge_ratio: Ratio of total asset value to hedge.
        :param hedge_leverage: Leverage to use for hedging.
        :param slippage: Slippage to use for hedging.
        :param max_order_age: Maximum age of an order before it is cancelled.
        :param min_trade_size: Minimum trade size.
        :param hedge_interval: Interval to check for hedging.
        :param value_mode: True if the strategy is in value mode, False otherwise.
        :param hedge_position_mode: Position mode (ONEWAY or HEDGE) to use for hedging.
        :param status_report_interval: Interval to report status.
        """
        super().__init__()
        self._hedge_market_pairs = hedge_market_pairs
        self._market_pairs = market_pairs
        self._hedge_ratio = config_map.hedge_ratio
        self._leverage = config_map.hedge_leverage
        self._position_mode = PositionMode.ONEWAY if config_map.hedge_position_mode == "ONEWAY" else PositionMode.HEDGE
        self._slippage = config_map.slippage
        self._min_trade_size = config_map.min_trade_size
        self._hedge_interval = config_map.hedge_interval
        self._value_mode = config_map.value_mode
        self._offsets = offsets
        self._status_report_interval = status_report_interval
        self._all_markets = self._hedge_market_pairs + self._market_pairs
        self._last_timestamp = 0
        self._all_markets_ready = False
        self._max_order_age = max_order_age
        self._status_messages = []
        self._last_report_timestamp = {}
        self._enable_auto_set_position_mode = enable_auto_set_position_mode
        if config_map.value_mode:
            self.hedge = self.hedge_by_value
            self._hedge_market_pair = hedge_market_pairs[0]
            self.logger().info(f"Hedge market pair: {self._hedge_market_pair}")
        else:
            self.hedge = self.hedge_by_amount
            self._market_pair_by_asset = self.get_market_pair_by_asset()
            self.logger().info(f"Market pair by asset: {self._market_pair_by_asset}")

        derivative_markets = AllConnectorSettings.get_derivative_names()
        self._derivatives_list = [
            market_pair for market_pair in self._all_markets if market_pair.market.name in derivative_markets
        ]
        self.get_order_candidates = (
            self.get_perpetual_order_candidates
            if self.is_derivative(self._hedge_market_pairs[0])
            else self.get_spot_order_candidates
        )

        all_markets = list(set([market_pair.market for market_pair in self._all_markets]))
        self.add_markets(all_markets)

    def get_market_pair_by_asset(self) -> Dict[MarketTradingPairTuple, List[MarketTradingPairTuple]]:
        """
        sort market pair belonging to the same market as hedge market together
        :return: market pair belonging to the same market as hedge market together
        """
        self.logger().info(f"Market pairs: {self._market_pairs}")
        return {
            hedge_pair: [
                market_pair for market_pair in self._market_pairs
                if market_pair.trading_pair.split("-")[0] == hedge_pair.trading_pair.split("-")[0]
            ]
            for hedge_pair in self._hedge_market_pairs
        }

    def is_derivative(self, market_pair: MarketTradingPairTuple) -> bool:
        """
        Check if the market is derivative.
        :param market_pair: Market pair to check.
        :return: True if the market is derivative, False otherwise.
        """
        return market_pair in self._derivatives_list

    def active_positions_df(self) -> pd.DataFrame:
        """
        Get the active positions of all markets.
        :return: The active positions of all markets.
        """
        columns = ["Connector", "Symbol", "Type", "Entry", "Amount", "Leverage"]
        data = []
        for market_pair in self._all_markets:
            if not self.is_derivative(market_pair):
                continue
            for position in self.get_positions(market_pair):
                if not position:
                    continue
                data.append(
                    [
                        market_pair.market.name,
                        position.trading_pair,
                        position.position_side.name,
                        position.entry_price,
                        position.amount,
                        position.leverage,
                    ]
                )
        return pd.DataFrame(data=data, columns=columns)

    def wallet_df(self) -> pd.DataFrame:
        """
        Processes the data required for wallet dataframe.
        :return: wallet dataframe
        """
        data = []
        columns = ["Connector", "Asset", "Price", "Amount", "Value"]

        def get_data(market_pair: MarketTradingPairTuple) -> List[Any]:
            market, trading_pair = market_pair.market, market_pair.trading_pair
            return [
                market.name,
                trading_pair,
                market_pair.get_mid_price(),
                f"{self.get_base_amount(market_pair):.6g}",
                f"{self.get_base_amount(market_pair) * market_pair.get_mid_price():.6g}",
            ]

        for market_pair in self._all_markets:
            data.append(get_data(market_pair))
        return pd.DataFrame(data=data, columns=columns)

    @property
    def active_orders(self) -> List[Tuple[Any, LimitOrder]]:
        """
        Get the active orders of all markets.
        :return: The active orders of all hedge markets.

        """
        return self.order_tracker.active_limit_orders

    def format_status(self) -> str:
        """
        Format the status of the strategy.
        """
        def get_wallet_status_str() -> List[str]:
            wallet_df = self.wallet_balance_data_frame(self._all_markets)
            return ["", "  Wallet:"] + ["    " + line for line in str(wallet_df).split("\n")]

        def get_asset_status_str() -> List[str]:
            assets_df = self.wallet_df()
            return ["", "  Assets:"] + ["    " + line for line in str(assets_df).split("\n")]

        def get_position_status_str() -> List[str]:
            positions_df = self.active_positions_df()
            if not positions_df.empty:
                return ["", "  Positions:"] + ["    " + line for line in str(positions_df).split("\n")]
            return ["", "  No positions."]

        def get_order_status_str() -> List[str]:
            if self.active_orders:
                orders = [order[1] for order in self.active_orders]
                df = LimitOrder.to_pandas(orders)
                df_lines = str(df).split("\n")
                return ["", "  Active orders:"] + ["    " + line for line in df_lines]
            return ["", "  No active maker orders."]

        def get_value_mode_status_str(value_mode: bool) -> List[str]:
            if not value_mode:
                return []

            lines = []
            total_value = sum(self.get_base_value(market_pair) for market_pair in self._market_pairs)
            hedge_value = self.get_base_value(self._hedge_market_pair)
            is_buy, value_to_hedge = self.get_hedge_direction_and_value()
            price, amount = self.calculate_hedge_price_and_amount(is_buy, value_to_hedge)
            lines.extend(["", f"   Mode: Value, Total value: {total_value:.6g}, Hedge value: {hedge_value:.6g}"])
            if amount > 0:
                lines.extend(
                    [
                        "",
                        f"   Next Hedge direction: {'buy' if is_buy else 'sell'}, Hedge price: {price:.6g}, Hedge amount: {amount:.6g}",
                    ]
                )
            return lines

        def get_amount_mode_status_str(value_mode: bool) -> List[str]:
            if value_mode:
                return []
            lines = ["", "   Mode: Amount"]
            data = []
            for hedge_market, market_list in self._market_pair_by_asset.items():
                hedge_market_name = hedge_market.market.name
                asset = hedge_market.trading_pair.split("-")[0]
                market_names = ", ".join([market.market.name for market in market_list])
                total_amount = sum(self.get_base_amount(market_pair) for market_pair in market_list)
                hedge_amount = self.get_base_amount(hedge_market)
                net_amount = total_amount * self._hedge_ratio + hedge_amount
                data.append([
                    hedge_market_name,
                    asset,
                    total_amount,
                    hedge_amount,
                    net_amount,
                    market_names,
                ])

            df = pd.DataFrame(data=data, columns=["Hedge Market", "Asset", "Total Amount", "Hedge Amount", "Net Amount", "Markets"])
            lines.extend(["    " + line for line in str(df).split("\n")])
            return lines

        def get_last_checked_seconds_str() -> List[str]:
            if self._last_timestamp < 1e9:
                return ["  Last checked: Not started."]
            return [f"  Last checked {self.current_timestamp - self._last_timestamp} seconds ago."]

        def get_status_messages() -> List[str]:
            if self._status_messages:
                return ["", "  Status Messages:"] + ["    " + line for line in self._status_messages]
            return []

        lines = (
            get_wallet_status_str()
            + get_asset_status_str()
            + get_position_status_str()
            + get_order_status_str()
            + get_value_mode_status_str(self._value_mode)
            + get_amount_mode_status_str(self._value_mode)
            + get_last_checked_seconds_str()
            + get_status_messages()
            + [""]
            + self.network_warning(self._all_markets)
        )
        return "\n".join(lines)

    def start(self, clock: Clock, timestamp: float) -> None:
        """
        Start the strategy.
        :param clock: Clock to use.
        :param timestamp: Current time.
        """
        self._last_timestamp = timestamp
        self.apply_initial_setting()

    def apply_initial_setting(self) -> None:
        """
        Check if the market is derivative, and if so, set the initial setting.
        """
        if not self.is_derivative(self._hedge_market_pairs[0]):
            return
        if not self._enable_auto_set_position_mode:
            logging.info("Auto set position mode is disabled.")
            return
        position_mode = "ONEWAY" if self._position_mode == PositionMode.ONEWAY else "HEDGE"
        msg = (
            f"Please ensure that the position mode on {self._hedge_market_pairs[0].market.name} "
            f"is set to {position_mode}. "
            f"The bot will try to automatically set position mode to {position_mode}. "
            f"You may ignore the message if the position mode is already set to {position_mode}.")
        self.notify_hb_app(msg)
        self.logger().warning(msg)
        for market_pair in self._hedge_market_pairs:
            market = market_pair.market
            trading_pair = market_pair.trading_pair
            market.set_leverage(trading_pair, self._leverage)
            market.set_position_mode(self._position_mode)

    def interval_log(self, key: str, message: str) -> None:
        """
        Log message at interval.
        :param key: Key to identify the last recorded timestamp.
        :param message: Message to log.
        """
        if self._last_timestamp - self._last_report_timestamp.get(key, 0) > self._status_report_interval:
            self.logger().info(message)
            self._last_report_timestamp[key] = self._last_timestamp

    def tick(self, timestamp: float) -> None:
        """
        Check if hedge interval has passed and process hedge if so
        :param timestamp: clock timestamp
        """
        if self.check_and_cancel_active_orders():
            self.interval_log("hedge", "Active orders present. Skipping hedge check until active orders expires.")
            return
        if timestamp - self._last_timestamp < self._hedge_interval:
            return
        self._all_markets_ready = all([market.ready for market in self.active_markets])
        if not self._all_markets_ready:
            # Markets not ready yet. Don't do anything.
            for market in self.active_markets:
                if not market.ready:
                    self.logger().warning(f"Market {market.name} is not ready.")
            self.logger().warning("Markets are not ready. No hedge trades are permitted.")
            return

        if not all([market.network_status is NetworkStatus.CONNECTED for market in self.active_markets]):
            self.logger().warning(
                "WARNING: Some markets are not connected or are down at the moment. "
                "Hedging may be dangerous when markets or networks are unstable. "
                "Retrying after %ss.",
                self._hedge_interval,
            )
            return
        self.interval_log("hedge", "Checking hedge conditions...")
        self._status_messages = []
        self.hedge()
        self._last_timestamp = timestamp

    def get_positions(self, market_pair: MarketTradingPairTuple, position_side: PositionSide = None) -> List[Position]:
        """
        Get the active positions of a market.
        :param market_pair: Market pair to get the positions of.
        :return: The active positions of the market.
        """
        trading_pair = market_pair.trading_pair
        positions: List[Position] = [
            position
            for position in market_pair.market.account_positions.values()
            if not isinstance(position, PositionMode) and position.trading_pair == trading_pair
        ]
        if position_side:
            return [position for position in positions if position.position_side == position_side]
        return positions

    def get_derivative_base_amount(self, market_pair: MarketTradingPairTuple) -> Decimal:
        """
        Get the value of the derivative base asset.
        :param market_pair: The market pair to get the value of the derivative base asset.
        :return: The value of the derivative base asset.
        """
        positions = self.get_positions(market_pair)
        amount = 0

        for position in positions:
            if position.position_side in [PositionSide.LONG, PositionSide.BOTH]:
                amount += position.amount
            if position.position_side == PositionSide.SHORT:
                amount -= abs(position.amount)
        return amount + self._offsets[market_pair]

    def get_base_amount(self, market_pair: MarketTradingPairTuple) -> Decimal:
        """
        Get the amount of the base asset of the market pair.

        :params market_pair: The market pair to get the amount of the base asset of.
        :returns: The amount of the base asset of the market pair.
        """
        if self.is_derivative(market_pair):
            self.logger().debug(f"Getting derivative base amount for {market_pair.trading_pair}")
            return self.get_derivative_base_amount(market_pair)
        return market_pair.base_balance + self._offsets[market_pair]

    def get_base_value(self, market_pair: MarketTradingPairTuple) -> Decimal:
        """
        Get the base asset value of a market. e.g BTC/USDT = BTC amount * BTC/USDT price.

        :params market_pair: The market pair to get the base asset value of.
        :returns: The base asset value of the market pair.
        """
        base_amount = self.get_base_amount(market_pair)
        base_price = market_pair.get_mid_price()
        return base_amount * base_price

    def get_hedge_direction_and_value(self) -> Tuple[bool, Decimal]:
        """
        Calculate the value that is required to be hedged.
        :returns: A tuple of the hedge direction (buy/sell) and the value to be hedged.
        """
        total_value = sum(self.get_base_value(market_pair) for market_pair in self._market_pairs)
        hedge_value = self.get_base_value(self._hedge_market_pair)
        net_value = total_value * self._hedge_ratio + hedge_value
        is_buy = net_value < 0
        value_to_hedge = abs(net_value)
        return is_buy, value_to_hedge

    def get_slippage_ratio(self, is_buy: bool) -> Decimal:
        """
        Get the slippage ratio for a buy or sell.
        :param is_buy: True if buy, False if sell.
        :returns: The ratio to multiply the price by to account for slippage.
        """
        return 1 + self._slippage if is_buy else 1 - self._slippage

    def calculate_hedge_price_and_amount(self, is_buy: bool, value_to_hedge: Decimal) -> Tuple[Decimal, Decimal]:
        """
        Calculate the price and amount to hedge.
        :params is_buy: The direction of the hedge.
        :params value_to_hedge: The value to hedge.
        :returns: The price and amount to hedge.
        """
        price = self._hedge_market_pair.get_mid_price()
        amount = value_to_hedge / price
        price = price * self.get_slippage_ratio(is_buy)
        trading_pair = self._hedge_market_pair.trading_pair
        quantized_price = self._hedge_market_pair.market.quantize_order_price(trading_pair, price)
        quantized_amount = self._hedge_market_pair.market.quantize_order_amount(trading_pair, amount)
        return quantized_price, quantized_amount

    def hedge_by_value(self) -> None:
        """
        The main process of the strategy for value mode = True.
        """
        is_buy, value_to_hedge = self.get_hedge_direction_and_value()
        price, amount = self.calculate_hedge_price_and_amount(is_buy, value_to_hedge)
        if amount == Decimal("0"):
            self.logger().debug("No hedge required.")
            self._status_messages.append("No hedge required.")
            return
        self.logger().info(
            f"Hedging by value. Hedge direction: {'buy' if is_buy else 'sell'}. "
            f"Hedge price: {price}. Hedge amount: {amount}."
        )
        order_candidates = self.get_order_candidates(self._hedge_market_pair, is_buy, amount, price)
        if not order_candidates:
            self.logger().info("No order candidates.")
            self._status_messages.append("No order candidates.")
            return
        self.place_orders(self._hedge_market_pair, order_candidates)

    def get_hedge_direction_and_amount_by_asset(
        self, hedge_pair: MarketTradingPairTuple, market_list: List[MarketTradingPairTuple]
    ) -> Tuple[bool, Decimal]:
        """
        Calculate the amount that is required to be hedged.
        :params hedge_pair: The market pair to hedge.
        :params market_list: The list of markets to get the amount of the base asset of.
        :returns: The direction to hedge (buy/sell) and the amount of the base asset of the market pair.
        """
        total_amount = 0
        for market_pair in market_list:
            amount = self.get_base_amount(market_pair)
            total_amount += self.get_base_amount(market_pair)
            self.logger().debug("Market pair: %s amount: %s, total_amount: %s", market_pair, amount, total_amount)

        hedge_amount = self.get_base_amount(hedge_pair)
        net_amount = total_amount * self._hedge_ratio + hedge_amount
        is_buy = net_amount < 0
        amount_to_hedge = abs(net_amount)
        self.logger().debug("Hedge direction: %s, amount to hedge: %s net amount: %s", is_buy, amount_to_hedge, net_amount)
        return is_buy, amount_to_hedge

    def hedge_by_amount(self) -> None:
        """
        The main process of the strategy for value mode = False.
        """
        for hedge_market, market_list in self._market_pair_by_asset.items():
            is_buy, amount_to_hedge = self.get_hedge_direction_and_amount_by_asset(hedge_market, market_list)
            asset = hedge_market.trading_pair.split("-")[0]
            self.logger().debug("Hedge by amount for %s: %s", asset, amount_to_hedge)
            self._status_messages.append(f"Hedge by amount for {asset}: {amount_to_hedge}")
            if amount_to_hedge == 0:
                self.logger().debug("No hedge required for %s.", asset)
                self._status_messages.append(f"No hedge required for {asset}.")
                continue
            price = hedge_market.get_mid_price() * self.get_slippage_ratio(is_buy)
            self.logger().info(
                "Hedge by amount. Mid price: %s Hedge direction: %s. Hedge price: %s. Hedge amount: %s",
                hedge_market.get_mid_price(), is_buy, price, amount_to_hedge
            )
            order_candidates = self.get_order_candidates(hedge_market, is_buy, amount_to_hedge, price)
            if not order_candidates:
                self.logger().info("Difference in hedge_amount found but no order candidates for %s is available. “This is either due to insufficient balance to perform hedge or min trade size not reached or minimum exchange trade size not met", asset)
                self._status_messages.append(f"No order candidates for {asset}.")
                continue
            self.place_orders(hedge_market, order_candidates)

    def get_perpetual_order_candidates(
        self, market_pair: MarketTradingPairTuple, is_buy: bool, amount: Decimal, price: Decimal
    ) -> List[PerpetualOrderCandidate]:
        """
        Check if the balance is sufficient to place an order.
        if not, adjust the amount to the balance available.
        returns the order candidate if the order meets the accepted criteria
        else, return None
        """
        self.logger().info("Checking perpetual order candidates for %s %s %s %s", market_pair, "buy" if is_buy else "sell", amount, price)

        def get_closing_order_candidate(is_buy: bool, amount: Decimal, price: Decimal) -> Union[PerpetualOrderCandidate, None]:
            opp_position_side = PositionSide.SHORT if is_buy else PositionSide.LONG
            opp_position_list = self.get_positions(market_pair, opp_position_side)
            # opp_position_list should only have 1 position
            for opp_position in opp_position_list:
                close_amount = min(amount, abs(opp_position.amount))
                order_candidate = PerpetualOrderCandidate(
                    trading_pair=market_pair.trading_pair,
                    is_maker=False,
                    order_side=TradeType.BUY if is_buy else TradeType.SELL,
                    amount=close_amount,
                    price=price,
                    order_type=OrderType.LIMIT,
                    leverage=Decimal(self._leverage),
                    position_close=True,
                )
                adjusted_candidate_order = budget_checker.adjust_candidate(order_candidate, all_or_none=False)
                return adjusted_candidate_order
            return None

        budget_checker = market_pair.market.budget_checker
        if amount * price < self._min_trade_size:
            self.logger().info("trade value (%s) is less than min trade size. (%s)", amount * price, self._min_trade_size)
            return []
        order_candidates = []
        if self._position_mode == PositionMode.HEDGE:
            order_candidate = get_closing_order_candidate(is_buy, amount, price)
            if order_candidate:
                order_candidates.append(order_candidate)
                amount -= order_candidate.amount
        order_candidate = PerpetualOrderCandidate(
            trading_pair=market_pair.trading_pair,
            is_maker=False,
            order_type=OrderType.LIMIT,
            order_side=TradeType.BUY if is_buy else TradeType.SELL,
            amount=amount,
            price=price,
            leverage=Decimal(self._leverage),
        )
        self.logger().info("order candidate: %s", order_candidate)
        adjusted_candidate_order = budget_checker.adjust_candidate(order_candidate, all_or_none=False)
        self.logger().info("adjusted order candidate: %s", adjusted_candidate_order)
        if adjusted_candidate_order.amount > 0:
            order_candidates.append(adjusted_candidate_order)
        return order_candidates

    def get_spot_order_candidates(
        self, market_pair: MarketTradingPairTuple, is_buy: bool, amount: Decimal, price: Decimal
    ) -> List[OrderCandidate]:
        """
        Check if the balance is sufficient to place an order.
        if not, adjust the amount to the balance available.
        returns the order candidate if the order meets the accepted criteria
        else, return None
        """
        budget_checker = market_pair.market.budget_checker
        if amount * price < self._min_trade_size:
            self.logger().info("trade value (%s) is less than min trade size. (%s)", amount * price, self._min_trade_size)
            return []
        order_candidate = OrderCandidate(
            trading_pair=market_pair.trading_pair,
            is_maker=False,
            order_type=OrderType.LIMIT,
            order_side=TradeType.BUY if is_buy else TradeType.SELL,
            amount=amount,
            price=price,
        )
        self.logger().info("order candidate: %s", order_candidate)
        adjusted_candidate_order = budget_checker.adjust_candidate(order_candidate, all_or_none=False)
        self.logger().info("adjusted order candidate: %s", adjusted_candidate_order)
        if adjusted_candidate_order.amount > 0:
            return [adjusted_candidate_order]
        return []

    def place_orders(
        self, market_pair: MarketTradingPairTuple, orders: Union[List[OrderCandidate], List[PerpetualOrderCandidate]]
    ) -> None:
        """
        Place an order refering the order candidates.
        :params market_pair: The market pair to place the order.
        :params orders: The list of orders to place.
        """
        self.logger().info("Placing %s orders", len(orders))
        for order in orders:
            self.logger().info(f"Create {order.order_side} {order.amount} {order.trading_pair} at {order.price}")
            is_buy = order.order_side == TradeType.BUY
            amount = order.amount
            price = order.price
            position_action = PositionAction.OPEN
            if isinstance(order, PerpetualOrderCandidate) and order.position_close:
                position_action = PositionAction.CLOSE
            trade = self.buy_with_specific_market if is_buy else self.sell_with_specific_market
            trade(market_pair, amount, order_type=OrderType.LIMIT, price=price, position_action=position_action)

    def check_and_cancel_active_orders(self) -> bool:
        """
        Check if there are any active orders and cancel them
        :return: True if there are active orders, False otherwise.
        """
        if not self.active_orders:
            return False
        for market_pair, order in self.active_orders:
            if order_age(order, self.current_timestamp) < self._max_order_age:
                continue
            self.logger().info(
                f"Cancel {'buy' if order.is_buy else 'sell'} {order.quantity} {order.trading_pair} at {order.price}"
            )
            market_pair.cancel(order.trading_pair, order.client_order_id)
        return True



import asyncio
import json
from typing import TYPE_CHECKING, Any, Dict, List, Optional

from hummingbot.connector.exchange.bitmart import (
    bitmart_constants as CONSTANTS,
    bitmart_utils as utils,
    bitmart_web_utils as web_utils,
)
from hummingbot.core.data_type.common import TradeType
from hummingbot.core.data_type.order_book_message import OrderBookMessage, OrderBookMessageType
from hummingbot.core.data_type.order_book_tracker_data_source import OrderBookTrackerDataSource
from hummingbot.core.web_assistant.connections.data_types import RESTMethod, WSJSONRequest
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory
from hummingbot.core.web_assistant.ws_assistant import WSAssistant
from hummingbot.logger import HummingbotLogger

if TYPE_CHECKING:
    from hummingbot.connector.exchange.bitmart.bitmart_exchange import BitmartExchange


class BitmartAPIOrderBookDataSource(OrderBookTrackerDataSource):

    _logger: Optional[HummingbotLogger] = None

    def __init__(self,
                 trading_pairs: List[str],
                 connector: 'BitmartExchange',
                 api_factory: WebAssistantsFactory):
        super().__init__(trading_pairs)
        self._connector: BitmartExchange = connector
        self._api_factory = api_factory

    async def get_last_traded_prices(self,
                                     trading_pairs: List[str],
                                     domain: Optional[str] = None) -> Dict[str, float]:
        return await self._connector.get_last_traded_prices(trading_pairs=trading_pairs)

    async def listen_for_order_book_diffs(self, ev_loop: asyncio.AbstractEventLoop, output: asyncio.Queue):
        """
        Bitmart only sends full snapshots, they never send diff messages. That is why this method is overwritten to
        do nothing.

        :param ev_loop: the event loop the method will run in
        :param output: a queue to add the created diff messages
        """
        pass

    async def listen_for_order_book_snapshots(self, ev_loop: asyncio.AbstractEventLoop, output: asyncio.Queue):
        """
        Bitmart sends always full snapshots through the depth channel. That is why they are processed here.

        :param ev_loop: the event loop the method will run in
        :param output: a queue to add the created diff messages
        """
        message_queue = self._message_queue[self._diff_messages_queue_key]
        while True:
            try:
                snapshot_event = await message_queue.get()
                await self._parse_order_book_snapshot_message(raw_message=snapshot_event, message_queue=output)

            except asyncio.CancelledError:
                raise
            except Exception:
                self.logger().exception("Unexpected error when processing public order book updates from exchange")

    async def _order_book_snapshot(self, trading_pair: str) -> OrderBookMessage:
        snapshot_response: Dict[str, Any] = await self._request_order_book_snapshot(trading_pair)
        snapshot_data: Dict[str, Any] = snapshot_response["data"]
        snapshot_timestamp: float = int(snapshot_data["ts"]) * 1e-3
        update_id: int = int(snapshot_data["ts"])

        order_book_message_content = {
            "trading_pair": trading_pair,
            "update_id": update_id,
            "bids": [(bid[0], bid[1]) for bid in snapshot_data["bids"]],
            "asks": [(ask[0], ask[1]) for ask in snapshot_data["asks"]],
        }
        snapshot_msg: OrderBookMessage = OrderBookMessage(
            OrderBookMessageType.SNAPSHOT,
            order_book_message_content,
            snapshot_timestamp)

        return snapshot_msg

    async def _request_order_book_snapshot(self, trading_pair: str) -> Dict[str, Any]:
        """
        Retrieves a copy of the full order book from the exchange, for a particular trading pair.

        :param trading_pair: the trading pair for which the order book will be retrieved

        :return: the response from the exchange (JSON dictionary)
        """
        params = {
            "symbol": await self._connector.exchange_symbol_associated_to_pair(trading_pair=trading_pair),
            "size": 200
        }

        rest_assistant = await self._api_factory.get_rest_assistant()
        data = await rest_assistant.execute_request(
            url=web_utils.public_rest_url(path_url=CONSTANTS.GET_ORDER_BOOK_PATH_URL),
            params=params,
            method=RESTMethod.GET,
            throttler_limit_id=CONSTANTS.GET_ORDER_BOOK_PATH_URL,
        )

        return data

    async def _parse_trade_message(self, raw_message: Dict[str, Any], message_queue: asyncio.Queue):
        trade_updates = raw_message["data"]

        for trade_data in trade_updates:
            trading_pair = await self._connector.trading_pair_associated_to_exchange_symbol(symbol=trade_data["symbol"])
            message_content = {
                "trade_id": int(trade_data["s_t"]),
                "trading_pair": trading_pair,
                "trade_type": float(TradeType.BUY.value) if trade_data["side"] == "buy" else float(
                    TradeType.SELL.value),
                "amount": trade_data["size"],
                "price": trade_data["price"]
            }
            trade_message: Optional[OrderBookMessage] = OrderBookMessage(
                message_type=OrderBookMessageType.TRADE,
                content=message_content,
                timestamp=int(trade_data["s_t"]))

            message_queue.put_nowait(trade_message)

    async def _parse_order_book_diff_message(self, raw_message: Dict[str, Any], message_queue: asyncio.Queue):
        # Bitmart never sends diff messages. This method will never be called
        pass

    async def _parse_order_book_snapshot_message(self, raw_message: Dict[str, Any], message_queue: asyncio.Queue):
        diff_updates: Dict[str, Any] = raw_message["data"]

        for diff_data in diff_updates:
            timestamp: float = int(diff_data["ms_t"]) * 1e-3
            update_id: int = int(diff_data["ms_t"])
            trading_pair = await self._connector.trading_pair_associated_to_exchange_symbol(
                symbol=diff_data["symbol"])

            order_book_message_content = {
                "trading_pair": trading_pair,
                "update_id": update_id,
                "bids": [(bid[0], bid[1]) for bid in diff_data["bids"]],
                "asks": [(ask[0], ask[1]) for ask in diff_data["asks"]],
            }
            diff_message: OrderBookMessage = OrderBookMessage(
                OrderBookMessageType.SNAPSHOT,
                order_book_message_content,
                timestamp)

            message_queue.put_nowait(diff_message)

    async def _subscribe_channels(self, ws: WSAssistant):
        try:
            symbols = [await self._connector.exchange_symbol_associated_to_pair(trading_pair=trading_pair)
                       for trading_pair in self._trading_pairs]

            payload = {
                "op": "subscribe",
                "args": [f"{CONSTANTS.PUBLIC_TRADE_CHANNEL_NAME}:{symbol}" for symbol in symbols]
            }
            subscribe_trade_request: WSJSONRequest = WSJSONRequest(payload=payload)

            payload = {
                "op": "subscribe",
                "args": [f"{CONSTANTS.PUBLIC_DEPTH_CHANNEL_NAME}:{symbol}" for symbol in symbols]
            }
            subscribe_orderbook_request: WSJSONRequest = WSJSONRequest(payload=payload)

            async with self._api_factory.throttler.execute_task(limit_id=CONSTANTS.WS_SUBSCRIBE):
                await ws.send(subscribe_trade_request)
            async with self._api_factory.throttler.execute_task(limit_id=CONSTANTS.WS_SUBSCRIBE):
                await ws.send(subscribe_orderbook_request)

            self.logger().info("Subscribed to public order book and trade channels...")
        except asyncio.CancelledError:
            raise
        except Exception:
            self.logger().exception("Unexpected error occurred subscribing to order book trading and delta streams...")
            raise

    async def _process_websocket_messages(self, websocket_assistant: WSAssistant):
        async for ws_response in websocket_assistant.iter_messages():
            data: Dict[str, Any] = ws_response.data
            decompressed_data = utils.decompress_ws_message(data)
            try:
                if isinstance(decompressed_data, str):
                    json_data = json.loads(decompressed_data)
                else:
                    json_data = decompressed_data
            except Exception:
                self.logger().warning(f"Invalid event message received through the order book data source "
                                      f"connection ({decompressed_data})")
                continue

            if "errorCode" in json_data or "errorMessage" in json_data:
                raise ValueError(f"Error message received in the order book data source: {json_data}")

            channel: str = self._channel_originating_message(event_message=json_data)
            if channel in [self._diff_messages_queue_key, self._trade_messages_queue_key]:
                self._message_queue[channel].put_nowait(json_data)

    def _channel_originating_message(self, event_message: Dict[str, Any]) -> str:
        channel = ""
        if "data" in event_message:
            event_channel = event_message["table"]
            if event_channel == CONSTANTS.PUBLIC_TRADE_CHANNEL_NAME:
                channel = self._trade_messages_queue_key
            if event_channel == CONSTANTS.PUBLIC_DEPTH_CHANNEL_NAME:
                channel = self._diff_messages_queue_key

        return channel

    async def _connected_websocket_assistant(self) -> WSAssistant:
        ws: WSAssistant = await self._api_factory.get_ws_assistant()
        async with self._api_factory.throttler.execute_task(limit_id=CONSTANTS.WS_CONNECT):
            await ws.connect(
                ws_url=CONSTANTS.WSS_PUBLIC_URL,
                ping_timeout=CONSTANTS.WS_PING_TIMEOUT)
        return ws



import threading
from decimal import Decimal
from typing import TYPE_CHECKING

from hummingbot.connector.utils import split_hb_trading_pair, validate_trading_pair
from hummingbot.core.rate_oracle.rate_oracle import RateOracle
from hummingbot.core.utils.async_utils import safe_ensure_future
from hummingbot.exceptions import OracleRateUnavailable

s_float_0 = float(0)
s_decimal_0 = Decimal("0")

if TYPE_CHECKING:
    from hummingbot.client.hummingbot_application import HummingbotApplication  # noqa: F401


class RateCommand:
    def rate(self,  # type: HummingbotApplication
             pair: str,
             token: str
             ):
        if threading.current_thread() != threading.main_thread():
            self.ev_loop.call_soon_threadsafe(self.trades)
            return
        if pair:
            safe_ensure_future(self.show_rate(pair))
        elif token:
            safe_ensure_future(self.show_token_value(token))

    async def show_rate(self,  # type: HummingbotApplication
                        pair: str,
                        ):
        if not validate_trading_pair(pair):
            self.notify(f"Invalid trading pair {pair}")
        else:
            try:
                msg = await self.oracle_rate_msg(pair)
            except OracleRateUnavailable:
                msg = "Rate is not available."
            self.notify(msg)

    async def oracle_rate_msg(self,  # type: HummingbotApplication
                              pair: str):
        if not validate_trading_pair(pair):
            self.notify(f"Invalid trading pair {pair}")
        else:
            pair = pair.upper().strip('\"').strip("'")
            rate = await RateOracle.get_instance().rate_async(pair)
            if rate is None:
                raise OracleRateUnavailable
            base, quote = split_hb_trading_pair(pair)
            return f"Source: {RateOracle.get_instance().source.name}\n1 {base} = {rate} {quote}"

    async def show_token_value(self,  # type: HummingbotApplication
                               token: str
                               ):
        if "-" in token:
            self.notify(f"Expected a single token but got a pair {token}")
        else:
            self.notify(f"Source: {RateOracle.get_instance().source.name}")
            rate = await RateOracle.get_instance().get_rate(base_token=token)
            if rate is None:
                self.notify("Rate is not available.")
                return
            global_token = self.client_config_map.global_token.global_token_name
            token_symbol = self.client_config_map.global_token.global_token_symbol
            self.notify(f"1 {token} = {token_symbol} {rate} {global_token}")



from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class Side(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    BID: _ClassVar[Side]
    ASK: _ClassVar[Side]

class TimeInForce(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    IMMEDIATE_OR_CANCEL: _ClassVar[TimeInForce]
    GOOD_FOR_SESSION: _ClassVar[TimeInForce]
    FILL_OR_KILL: _ClassVar[TimeInForce]

class OrderType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    LIMIT: _ClassVar[OrderType]
    MARKET_LIMIT: _ClassVar[OrderType]
    MARKET_WITH_PROTECTION: _ClassVar[OrderType]

class SelfTradePrevention(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    CANCEL_RESTING: _ClassVar[SelfTradePrevention]
    CANCEL_AGGRESSING: _ClassVar[SelfTradePrevention]
    ALLOW_SELF_TRADE: _ClassVar[SelfTradePrevention]

class PostOnly(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    DISABLED: _ClassVar[PostOnly]
    ENABLED: _ClassVar[PostOnly]
BID: Side
ASK: Side
IMMEDIATE_OR_CANCEL: TimeInForce
GOOD_FOR_SESSION: TimeInForce
FILL_OR_KILL: TimeInForce
LIMIT: OrderType
MARKET_LIMIT: OrderType
MARKET_WITH_PROTECTION: OrderType
CANCEL_RESTING: SelfTradePrevention
CANCEL_AGGRESSING: SelfTradePrevention
ALLOW_SELF_TRADE: SelfTradePrevention
DISABLED: PostOnly
ENABLED: PostOnly

class Credentials(_message.Message):
    __slots__ = ("access_key_id", "signature", "timestamp")
    ACCESS_KEY_ID_FIELD_NUMBER: _ClassVar[int]
    SIGNATURE_FIELD_NUMBER: _ClassVar[int]
    TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    access_key_id: str
    signature: str
    timestamp: int
    def __init__(self, access_key_id: _Optional[str] = ..., signature: _Optional[str] = ..., timestamp: _Optional[int] = ...) -> None: ...

class OrderRequest(_message.Message):
    __slots__ = ("new", "cancel", "modify", "heartbeat", "mc")
    NEW_FIELD_NUMBER: _ClassVar[int]
    CANCEL_FIELD_NUMBER: _ClassVar[int]
    MODIFY_FIELD_NUMBER: _ClassVar[int]
    HEARTBEAT_FIELD_NUMBER: _ClassVar[int]
    MC_FIELD_NUMBER: _ClassVar[int]
    new: NewOrder
    cancel: CancelOrder
    modify: ModifyOrder
    heartbeat: Heartbeat
    mc: MassCancel
    def __init__(self, new: _Optional[_Union[NewOrder, _Mapping]] = ..., cancel: _Optional[_Union[CancelOrder, _Mapping]] = ..., modify: _Optional[_Union[ModifyOrder, _Mapping]] = ..., heartbeat: _Optional[_Union[Heartbeat, _Mapping]] = ..., mc: _Optional[_Union[MassCancel, _Mapping]] = ...) -> None: ...

class NewOrder(_message.Message):
    __slots__ = ("client_order_id", "request_id", "market_id", "price", "quantity", "side", "time_in_force", "order_type", "subaccount_id", "self_trade_prevention", "post_only", "cancel_on_disconnect")
    CLIENT_ORDER_ID_FIELD_NUMBER: _ClassVar[int]
    REQUEST_ID_FIELD_NUMBER: _ClassVar[int]
    MARKET_ID_FIELD_NUMBER: _ClassVar[int]
    PRICE_FIELD_NUMBER: _ClassVar[int]
    QUANTITY_FIELD_NUMBER: _ClassVar[int]
    SIDE_FIELD_NUMBER: _ClassVar[int]
    TIME_IN_FORCE_FIELD_NUMBER: _ClassVar[int]
    ORDER_TYPE_FIELD_NUMBER: _ClassVar[int]
    SUBACCOUNT_ID_FIELD_NUMBER: _ClassVar[int]
    SELF_TRADE_PREVENTION_FIELD_NUMBER: _ClassVar[int]
    POST_ONLY_FIELD_NUMBER: _ClassVar[int]
    CANCEL_ON_DISCONNECT_FIELD_NUMBER: _ClassVar[int]
    client_order_id: int
    request_id: int
    market_id: int
    price: int
    quantity: int
    side: Side
    time_in_force: TimeInForce
    order_type: OrderType
    subaccount_id: int
    self_trade_prevention: SelfTradePrevention
    post_only: PostOnly
    cancel_on_disconnect: bool
    def __init__(self, client_order_id: _Optional[int] = ..., request_id: _Optional[int] = ..., market_id: _Optional[int] = ..., price: _Optional[int] = ..., quantity: _Optional[int] = ..., side: _Optional[_Union[Side, str]] = ..., time_in_force: _Optional[_Union[TimeInForce, str]] = ..., order_type: _Optional[_Union[OrderType, str]] = ..., subaccount_id: _Optional[int] = ..., self_trade_prevention: _Optional[_Union[SelfTradePrevention, str]] = ..., post_only: _Optional[_Union[PostOnly, str]] = ..., cancel_on_disconnect: bool = ...) -> None: ...

class CancelOrder(_message.Message):
    __slots__ = ("market_id", "client_order_id", "request_id", "subaccount_id")
    MARKET_ID_FIELD_NUMBER: _ClassVar[int]
    CLIENT_ORDER_ID_FIELD_NUMBER: _ClassVar[int]
    REQUEST_ID_FIELD_NUMBER: _ClassVar[int]
    SUBACCOUNT_ID_FIELD_NUMBER: _ClassVar[int]
    market_id: int
    client_order_id: int
    request_id: int
    subaccount_id: int
    def __init__(self, market_id: _Optional[int] = ..., client_order_id: _Optional[int] = ..., request_id: _Optional[int] = ..., subaccount_id: _Optional[int] = ...) -> None: ...

class ModifyOrder(_message.Message):
    __slots__ = ("market_id", "client_order_id", "request_id", "new_price", "new_quantity", "subaccount_id", "self_trade_prevention", "post_only")
    MARKET_ID_FIELD_NUMBER: _ClassVar[int]
    CLIENT_ORDER_ID_FIELD_NUMBER: _ClassVar[int]
    REQUEST_ID_FIELD_NUMBER: _ClassVar[int]
    NEW_PRICE_FIELD_NUMBER: _ClassVar[int]
    NEW_QUANTITY_FIELD_NUMBER: _ClassVar[int]
    SUBACCOUNT_ID_FIELD_NUMBER: _ClassVar[int]
    SELF_TRADE_PREVENTION_FIELD_NUMBER: _ClassVar[int]
    POST_ONLY_FIELD_NUMBER: _ClassVar[int]
    market_id: int
    client_order_id: int
    request_id: int
    new_price: int
    new_quantity: int
    subaccount_id: int
    self_trade_prevention: SelfTradePrevention
    post_only: PostOnly
    def __init__(self, market_id: _Optional[int] = ..., client_order_id: _Optional[int] = ..., request_id: _Optional[int] = ..., new_price: _Optional[int] = ..., new_quantity: _Optional[int] = ..., subaccount_id: _Optional[int] = ..., self_trade_prevention: _Optional[_Union[SelfTradePrevention, str]] = ..., post_only: _Optional[_Union[PostOnly, str]] = ...) -> None: ...

class MassCancel(_message.Message):
    __slots__ = ("subaccount_id", "request_id", "market_id", "side")
    SUBACCOUNT_ID_FIELD_NUMBER: _ClassVar[int]
    REQUEST_ID_FIELD_NUMBER: _ClassVar[int]
    MARKET_ID_FIELD_NUMBER: _ClassVar[int]
    SIDE_FIELD_NUMBER: _ClassVar[int]
    subaccount_id: int
    request_id: int
    market_id: int
    side: Side
    def __init__(self, subaccount_id: _Optional[int] = ..., request_id: _Optional[int] = ..., market_id: _Optional[int] = ..., side: _Optional[_Union[Side, str]] = ...) -> None: ...

class Heartbeat(_message.Message):
    __slots__ = ("request_id", "timestamp")
    REQUEST_ID_FIELD_NUMBER: _ClassVar[int]
    TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    request_id: int
    timestamp: int
    def __init__(self, request_id: _Optional[int] = ..., timestamp: _Optional[int] = ...) -> None: ...

class OrderResponse(_message.Message):
    __slots__ = ("new_ack", "cancel_ack", "modify_ack", "new_reject", "cancel_reject", "modify_reject", "fill", "heartbeat", "position", "mass_cancel_ack")
    NEW_ACK_FIELD_NUMBER: _ClassVar[int]
    CANCEL_ACK_FIELD_NUMBER: _ClassVar[int]
    MODIFY_ACK_FIELD_NUMBER: _ClassVar[int]
    NEW_REJECT_FIELD_NUMBER: _ClassVar[int]
    CANCEL_REJECT_FIELD_NUMBER: _ClassVar[int]
    MODIFY_REJECT_FIELD_NUMBER: _ClassVar[int]
    FILL_FIELD_NUMBER: _ClassVar[int]
    HEARTBEAT_FIELD_NUMBER: _ClassVar[int]
    POSITION_FIELD_NUMBER: _ClassVar[int]
    MASS_CANCEL_ACK_FIELD_NUMBER: _ClassVar[int]
    new_ack: NewOrderAck
    cancel_ack: CancelOrderAck
    modify_ack: ModifyOrderAck
    new_reject: NewOrderReject
    cancel_reject: CancelOrderReject
    modify_reject: ModifyOrderReject
    fill: Fill
    heartbeat: Heartbeat
    position: AssetPosition
    mass_cancel_ack: MassCancelAck
    def __init__(self, new_ack: _Optional[_Union[NewOrderAck, _Mapping]] = ..., cancel_ack: _Optional[_Union[CancelOrderAck, _Mapping]] = ..., modify_ack: _Optional[_Union[ModifyOrderAck, _Mapping]] = ..., new_reject: _Optional[_Union[NewOrderReject, _Mapping]] = ..., cancel_reject: _Optional[_Union[CancelOrderReject, _Mapping]] = ..., modify_reject: _Optional[_Union[ModifyOrderReject, _Mapping]] = ..., fill: _Optional[_Union[Fill, _Mapping]] = ..., heartbeat: _Optional[_Union[Heartbeat, _Mapping]] = ..., position: _Optional[_Union[AssetPosition, _Mapping]] = ..., mass_cancel_ack: _Optional[_Union[MassCancelAck, _Mapping]] = ...) -> None: ...

class NewOrderAck(_message.Message):
    __slots__ = ("msg_seq_num", "client_order_id", "request_id", "exchange_order_id", "market_id", "price", "quantity", "side", "time_in_force", "order_type", "transact_time", "subaccount_id", "cancel_on_disconnect")
    MSG_SEQ_NUM_FIELD_NUMBER: _ClassVar[int]
    CLIENT_ORDER_ID_FIELD_NUMBER: _ClassVar[int]
    REQUEST_ID_FIELD_NUMBER: _ClassVar[int]
    EXCHANGE_ORDER_ID_FIELD_NUMBER: _ClassVar[int]
    MARKET_ID_FIELD_NUMBER: _ClassVar[int]
    PRICE_FIELD_NUMBER: _ClassVar[int]
    QUANTITY_FIELD_NUMBER: _ClassVar[int]
    SIDE_FIELD_NUMBER: _ClassVar[int]
    TIME_IN_FORCE_FIELD_NUMBER: _ClassVar[int]
    ORDER_TYPE_FIELD_NUMBER: _ClassVar[int]
    TRANSACT_TIME_FIELD_NUMBER: _ClassVar[int]
    SUBACCOUNT_ID_FIELD_NUMBER: _ClassVar[int]
    CANCEL_ON_DISCONNECT_FIELD_NUMBER: _ClassVar[int]
    msg_seq_num: int
    client_order_id: int
    request_id: int
    exchange_order_id: int
    market_id: int
    price: int
    quantity: int
    side: Side
    time_in_force: TimeInForce
    order_type: OrderType
    transact_time: int
    subaccount_id: int
    cancel_on_disconnect: bool
    def __init__(self, msg_seq_num: _Optional[int] = ..., client_order_id: _Optional[int] = ..., request_id: _Optional[int] = ..., exchange_order_id: _Optional[int] = ..., market_id: _Optional[int] = ..., price: _Optional[int] = ..., quantity: _Optional[int] = ..., side: _Optional[_Union[Side, str]] = ..., time_in_force: _Optional[_Union[TimeInForce, str]] = ..., order_type: _Optional[_Union[OrderType, str]] = ..., transact_time: _Optional[int] = ..., subaccount_id: _Optional[int] = ..., cancel_on_disconnect: bool = ...) -> None: ...

class CancelOrderAck(_message.Message):
    __slots__ = ("msg_seq_num", "client_order_id", "request_id", "transact_time", "subaccount_id", "reason", "market_id", "exchange_order_id")
    class Reason(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        UNCLASSIFIED: _ClassVar[CancelOrderAck.Reason]
        DISCONNECT: _ClassVar[CancelOrderAck.Reason]
        REQUESTED: _ClassVar[CancelOrderAck.Reason]
        IOC: _ClassVar[CancelOrderAck.Reason]
        STP_RESTING: _ClassVar[CancelOrderAck.Reason]
        STP_AGGRESSING: _ClassVar[CancelOrderAck.Reason]
        MASS_CANCEL: _ClassVar[CancelOrderAck.Reason]
        POSITION_LIMIT: _ClassVar[CancelOrderAck.Reason]
    UNCLASSIFIED: CancelOrderAck.Reason
    DISCONNECT: CancelOrderAck.Reason
    REQUESTED: CancelOrderAck.Reason
    IOC: CancelOrderAck.Reason
    STP_RESTING: CancelOrderAck.Reason
    STP_AGGRESSING: CancelOrderAck.Reason
    MASS_CANCEL: CancelOrderAck.Reason
    POSITION_LIMIT: CancelOrderAck.Reason
    MSG_SEQ_NUM_FIELD_NUMBER: _ClassVar[int]
    CLIENT_ORDER_ID_FIELD_NUMBER: _ClassVar[int]
    REQUEST_ID_FIELD_NUMBER: _ClassVar[int]
    TRANSACT_TIME_FIELD_NUMBER: _ClassVar[int]
    SUBACCOUNT_ID_FIELD_NUMBER: _ClassVar[int]
    REASON_FIELD_NUMBER: _ClassVar[int]
    MARKET_ID_FIELD_NUMBER: _ClassVar[int]
    EXCHANGE_ORDER_ID_FIELD_NUMBER: _ClassVar[int]
    msg_seq_num: int
    client_order_id: int
    request_id: int
    transact_time: int
    subaccount_id: int
    reason: CancelOrderAck.Reason
    market_id: int
    exchange_order_id: int
    def __init__(self, msg_seq_num: _Optional[int] = ..., client_order_id: _Optional[int] = ..., request_id: _Optional[int] = ..., transact_time: _Optional[int] = ..., subaccount_id: _Optional[int] = ..., reason: _Optional[_Union[CancelOrderAck.Reason, str]] = ..., market_id: _Optional[int] = ..., exchange_order_id: _Optional[int] = ...) -> None: ...

class ModifyOrderAck(_message.Message):
    __slots__ = ("msg_seq_num", "client_order_id", "request_id", "transact_time", "remaining_quantity", "subaccount_id", "market_id", "price", "quantity", "cumulative_quantity", "exchange_order_id")
    MSG_SEQ_NUM_FIELD_NUMBER: _ClassVar[int]
    CLIENT_ORDER_ID_FIELD_NUMBER: _ClassVar[int]
    REQUEST_ID_FIELD_NUMBER: _ClassVar[int]
    TRANSACT_TIME_FIELD_NUMBER: _ClassVar[int]
    REMAINING_QUANTITY_FIELD_NUMBER: _ClassVar[int]
    SUBACCOUNT_ID_FIELD_NUMBER: _ClassVar[int]
    MARKET_ID_FIELD_NUMBER: _ClassVar[int]
    PRICE_FIELD_NUMBER: _ClassVar[int]
    QUANTITY_FIELD_NUMBER: _ClassVar[int]
    CUMULATIVE_QUANTITY_FIELD_NUMBER: _ClassVar[int]
    EXCHANGE_ORDER_ID_FIELD_NUMBER: _ClassVar[int]
    msg_seq_num: int
    client_order_id: int
    request_id: int
    transact_time: int
    remaining_quantity: int
    subaccount_id: int
    market_id: int
    price: int
    quantity: int
    cumulative_quantity: int
    exchange_order_id: int
    def __init__(self, msg_seq_num: _Optional[int] = ..., client_order_id: _Optional[int] = ..., request_id: _Optional[int] = ..., transact_time: _Optional[int] = ..., remaining_quantity: _Optional[int] = ..., subaccount_id: _Optional[int] = ..., market_id: _Optional[int] = ..., price: _Optional[int] = ..., quantity: _Optional[int] = ..., cumulative_quantity: _Optional[int] = ..., exchange_order_id: _Optional[int] = ...) -> None: ...

class MassCancelAck(_message.Message):
    __slots__ = ("msg_seq_num", "subaccount_id", "request_id", "transact_time", "reason", "total_affected_orders")
    class Reason(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        UNCLASSIFIED: _ClassVar[MassCancelAck.Reason]
        INVALID_MARKET_ID: _ClassVar[MassCancelAck.Reason]
        INVALID_SIDE: _ClassVar[MassCancelAck.Reason]
    UNCLASSIFIED: MassCancelAck.Reason
    INVALID_MARKET_ID: MassCancelAck.Reason
    INVALID_SIDE: MassCancelAck.Reason
    MSG_SEQ_NUM_FIELD_NUMBER: _ClassVar[int]
    SUBACCOUNT_ID_FIELD_NUMBER: _ClassVar[int]
    REQUEST_ID_FIELD_NUMBER: _ClassVar[int]
    TRANSACT_TIME_FIELD_NUMBER: _ClassVar[int]
    REASON_FIELD_NUMBER: _ClassVar[int]
    TOTAL_AFFECTED_ORDERS_FIELD_NUMBER: _ClassVar[int]
    msg_seq_num: int
    subaccount_id: int
    request_id: int
    transact_time: int
    reason: MassCancelAck.Reason
    total_affected_orders: int
    def __init__(self, msg_seq_num: _Optional[int] = ..., subaccount_id: _Optional[int] = ..., request_id: _Optional[int] = ..., transact_time: _Optional[int] = ..., reason: _Optional[_Union[MassCancelAck.Reason, str]] = ..., total_affected_orders: _Optional[int] = ...) -> None: ...

class NewOrderReject(_message.Message):
    __slots__ = ("msg_seq_num", "client_order_id", "request_id", "transact_time", "subaccount_id", "reason", "market_id", "price", "quantity", "side", "time_in_force", "order_type")
    class Reason(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        UNCLASSIFIED: _ClassVar[NewOrderReject.Reason]
        INVALID_QUANTITY: _ClassVar[NewOrderReject.Reason]
        INVALID_MARKET_ID: _ClassVar[NewOrderReject.Reason]
        DUPLICATE_ORDER_ID: _ClassVar[NewOrderReject.Reason]
        INVALID_SIDE: _ClassVar[NewOrderReject.Reason]
        INVALID_TIME_IN_FORCE: _ClassVar[NewOrderReject.Reason]
        INVALID_ORDER_TYPE: _ClassVar[NewOrderReject.Reason]
        INVALID_POST_ONLY: _ClassVar[NewOrderReject.Reason]
        INVALID_SELF_TRADE_PREVENTION: _ClassVar[NewOrderReject.Reason]
        UNKNOWN_TRADER: _ClassVar[NewOrderReject.Reason]
        PRICE_WITH_MARKET_LIMIT_ORDER: _ClassVar[NewOrderReject.Reason]
        POST_ONLY_WITH_MARKET_ORDER: _ClassVar[NewOrderReject.Reason]
        POST_ONLY_WITH_INVALID_TIF: _ClassVar[NewOrderReject.Reason]
        EXCEEDED_SPOT_POSITION: _ClassVar[NewOrderReject.Reason]
        NO_OPPOSING_RESTING_ORDER: _ClassVar[NewOrderReject.Reason]
        POST_ONLY_WOULD_TRADE: _ClassVar[NewOrderReject.Reason]
        DID_NOT_FULLY_FILL: _ClassVar[NewOrderReject.Reason]
        ONLY_ORDER_CANCEL_ACCEPTED: _ClassVar[NewOrderReject.Reason]
        PROTECTION_PRICE_WOULD_NOT_TRADE: _ClassVar[NewOrderReject.Reason]
        NO_REFERENCE_PRICE: _ClassVar[NewOrderReject.Reason]
        SLIPPAGE_TOO_HIGH: _ClassVar[NewOrderReject.Reason]
        OUTSIDE_PRICE_BAND: _ClassVar[NewOrderReject.Reason]
    UNCLASSIFIED: NewOrderReject.Reason
    INVALID_QUANTITY: NewOrderReject.Reason
    INVALID_MARKET_ID: NewOrderReject.Reason
    DUPLICATE_ORDER_ID: NewOrderReject.Reason
    INVALID_SIDE: NewOrderReject.Reason
    INVALID_TIME_IN_FORCE: NewOrderReject.Reason
    INVALID_ORDER_TYPE: NewOrderReject.Reason
    INVALID_POST_ONLY: NewOrderReject.Reason
    INVALID_SELF_TRADE_PREVENTION: NewOrderReject.Reason
    UNKNOWN_TRADER: NewOrderReject.Reason
    PRICE_WITH_MARKET_LIMIT_ORDER: NewOrderReject.Reason
    POST_ONLY_WITH_MARKET_ORDER: NewOrderReject.Reason
    POST_ONLY_WITH_INVALID_TIF: NewOrderReject.Reason
    EXCEEDED_SPOT_POSITION: NewOrderReject.Reason
    NO_OPPOSING_RESTING_ORDER: NewOrderReject.Reason
    POST_ONLY_WOULD_TRADE: NewOrderReject.Reason
    DID_NOT_FULLY_FILL: NewOrderReject.Reason
    ONLY_ORDER_CANCEL_ACCEPTED: NewOrderReject.Reason
    PROTECTION_PRICE_WOULD_NOT_TRADE: NewOrderReject.Reason
    NO_REFERENCE_PRICE: NewOrderReject.Reason
    SLIPPAGE_TOO_HIGH: NewOrderReject.Reason
    OUTSIDE_PRICE_BAND: NewOrderReject.Reason
    MSG_SEQ_NUM_FIELD_NUMBER: _ClassVar[int]
    CLIENT_ORDER_ID_FIELD_NUMBER: _ClassVar[int]
    REQUEST_ID_FIELD_NUMBER: _ClassVar[int]
    TRANSACT_TIME_FIELD_NUMBER: _ClassVar[int]
    SUBACCOUNT_ID_FIELD_NUMBER: _ClassVar[int]
    REASON_FIELD_NUMBER: _ClassVar[int]
    MARKET_ID_FIELD_NUMBER: _ClassVar[int]
    PRICE_FIELD_NUMBER: _ClassVar[int]
    QUANTITY_FIELD_NUMBER: _ClassVar[int]
    SIDE_FIELD_NUMBER: _ClassVar[int]
    TIME_IN_FORCE_FIELD_NUMBER: _ClassVar[int]
    ORDER_TYPE_FIELD_NUMBER: _ClassVar[int]
    msg_seq_num: int
    client_order_id: int
    request_id: int
    transact_time: int
    subaccount_id: int
    reason: NewOrderReject.Reason
    market_id: int
    price: int
    quantity: int
    side: Side
    time_in_force: TimeInForce
    order_type: OrderType
    def __init__(self, msg_seq_num: _Optional[int] = ..., client_order_id: _Optional[int] = ..., request_id: _Optional[int] = ..., transact_time: _Optional[int] = ..., subaccount_id: _Optional[int] = ..., reason: _Optional[_Union[NewOrderReject.Reason, str]] = ..., market_id: _Optional[int] = ..., price: _Optional[int] = ..., quantity: _Optional[int] = ..., side: _Optional[_Union[Side, str]] = ..., time_in_force: _Optional[_Union[TimeInForce, str]] = ..., order_type: _Optional[_Union[OrderType, str]] = ...) -> None: ...

class CancelOrderReject(_message.Message):
    __slots__ = ("msg_seq_num", "client_order_id", "request_id", "transact_time", "subaccount_id", "reason", "market_id")
    class Reason(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        UNCLASSIFIED: _ClassVar[CancelOrderReject.Reason]
        INVALID_MARKET_ID: _ClassVar[CancelOrderReject.Reason]
        ORDER_NOT_FOUND: _ClassVar[CancelOrderReject.Reason]
    UNCLASSIFIED: CancelOrderReject.Reason
    INVALID_MARKET_ID: CancelOrderReject.Reason
    ORDER_NOT_FOUND: CancelOrderReject.Reason
    MSG_SEQ_NUM_FIELD_NUMBER: _ClassVar[int]
    CLIENT_ORDER_ID_FIELD_NUMBER: _ClassVar[int]
    REQUEST_ID_FIELD_NUMBER: _ClassVar[int]
    TRANSACT_TIME_FIELD_NUMBER: _ClassVar[int]
    SUBACCOUNT_ID_FIELD_NUMBER: _ClassVar[int]
    REASON_FIELD_NUMBER: _ClassVar[int]
    MARKET_ID_FIELD_NUMBER: _ClassVar[int]
    msg_seq_num: int
    client_order_id: int
    request_id: int
    transact_time: int
    subaccount_id: int
    reason: CancelOrderReject.Reason
    market_id: int
    def __init__(self, msg_seq_num: _Optional[int] = ..., client_order_id: _Optional[int] = ..., request_id: _Optional[int] = ..., transact_time: _Optional[int] = ..., subaccount_id: _Optional[int] = ..., reason: _Optional[_Union[CancelOrderReject.Reason, str]] = ..., market_id: _Optional[int] = ...) -> None: ...

class ModifyOrderReject(_message.Message):
    __slots__ = ("msg_seq_num", "client_order_id", "request_id", "transact_time", "subaccount_id", "reason", "market_id")
    class Reason(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        UNCLASSIFIED: _ClassVar[ModifyOrderReject.Reason]
        INVALID_QUANTITY: _ClassVar[ModifyOrderReject.Reason]
        INVALID_MARKET_ID: _ClassVar[ModifyOrderReject.Reason]
        ORDER_NOT_FOUND: _ClassVar[ModifyOrderReject.Reason]
        INVALID_IFM: _ClassVar[ModifyOrderReject.Reason]
        INVALID_POST_ONLY: _ClassVar[ModifyOrderReject.Reason]
        INVALID_SELF_TRADE_PREVENTION: _ClassVar[ModifyOrderReject.Reason]
        UNKNOWN_TRADER: _ClassVar[ModifyOrderReject.Reason]
        EXCEEDED_SPOT_POSITION: _ClassVar[ModifyOrderReject.Reason]
        POST_ONLY_WOULD_TRADE: _ClassVar[ModifyOrderReject.Reason]
        ONLY_ORDER_CANCEL_ACCEPTED: _ClassVar[ModifyOrderReject.Reason]
        OUTSIDE_PRICE_BAND: _ClassVar[ModifyOrderReject.Reason]
    UNCLASSIFIED: ModifyOrderReject.Reason
    INVALID_QUANTITY: ModifyOrderReject.Reason
    INVALID_MARKET_ID: ModifyOrderReject.Reason
    ORDER_NOT_FOUND: ModifyOrderReject.Reason
    INVALID_IFM: ModifyOrderReject.Reason
    INVALID_POST_ONLY: ModifyOrderReject.Reason
    INVALID_SELF_TRADE_PREVENTION: ModifyOrderReject.Reason
    UNKNOWN_TRADER: ModifyOrderReject.Reason
    EXCEEDED_SPOT_POSITION: ModifyOrderReject.Reason
    POST_ONLY_WOULD_TRADE: ModifyOrderReject.Reason
    ONLY_ORDER_CANCEL_ACCEPTED: ModifyOrderReject.Reason
    OUTSIDE_PRICE_BAND: ModifyOrderReject.Reason
    MSG_SEQ_NUM_FIELD_NUMBER: _ClassVar[int]
    CLIENT_ORDER_ID_FIELD_NUMBER: _ClassVar[int]
    REQUEST_ID_FIELD_NUMBER: _ClassVar[int]
    TRANSACT_TIME_FIELD_NUMBER: _ClassVar[int]
    SUBACCOUNT_ID_FIELD_NUMBER: _ClassVar[int]
    REASON_FIELD_NUMBER: _ClassVar[int]
    MARKET_ID_FIELD_NUMBER: _ClassVar[int]
    msg_seq_num: int
    client_order_id: int
    request_id: int
    transact_time: int
    subaccount_id: int
    reason: ModifyOrderReject.Reason
    market_id: int
    def __init__(self, msg_seq_num: _Optional[int] = ..., client_order_id: _Optional[int] = ..., request_id: _Optional[int] = ..., transact_time: _Optional[int] = ..., subaccount_id: _Optional[int] = ..., reason: _Optional[_Union[ModifyOrderReject.Reason, str]] = ..., market_id: _Optional[int] = ...) -> None: ...

class Fill(_message.Message):
    __slots__ = ("msg_seq_num", "market_id", "client_order_id", "exchange_order_id", "fill_price", "fill_quantity", "leaves_quantity", "transact_time", "subaccount_id", "cumulative_quantity", "side", "aggressor_indicator", "fee_ratio", "trade_id")
    MSG_SEQ_NUM_FIELD_NUMBER: _ClassVar[int]
    MARKET_ID_FIELD_NUMBER: _ClassVar[int]
    CLIENT_ORDER_ID_FIELD_NUMBER: _ClassVar[int]
    EXCHANGE_ORDER_ID_FIELD_NUMBER: _ClassVar[int]
    FILL_PRICE_FIELD_NUMBER: _ClassVar[int]
    FILL_QUANTITY_FIELD_NUMBER: _ClassVar[int]
    LEAVES_QUANTITY_FIELD_NUMBER: _ClassVar[int]
    TRANSACT_TIME_FIELD_NUMBER: _ClassVar[int]
    SUBACCOUNT_ID_FIELD_NUMBER: _ClassVar[int]
    CUMULATIVE_QUANTITY_FIELD_NUMBER: _ClassVar[int]
    SIDE_FIELD_NUMBER: _ClassVar[int]
    AGGRESSOR_INDICATOR_FIELD_NUMBER: _ClassVar[int]
    FEE_RATIO_FIELD_NUMBER: _ClassVar[int]
    TRADE_ID_FIELD_NUMBER: _ClassVar[int]
    msg_seq_num: int
    market_id: int
    client_order_id: int
    exchange_order_id: int
    fill_price: int
    fill_quantity: int
    leaves_quantity: int
    transact_time: int
    subaccount_id: int
    cumulative_quantity: int
    side: Side
    aggressor_indicator: bool
    fee_ratio: FixedPointDecimal
    trade_id: int
    def __init__(self, msg_seq_num: _Optional[int] = ..., market_id: _Optional[int] = ..., client_order_id: _Optional[int] = ..., exchange_order_id: _Optional[int] = ..., fill_price: _Optional[int] = ..., fill_quantity: _Optional[int] = ..., leaves_quantity: _Optional[int] = ..., transact_time: _Optional[int] = ..., subaccount_id: _Optional[int] = ..., cumulative_quantity: _Optional[int] = ..., side: _Optional[_Union[Side, str]] = ..., aggressor_indicator: bool = ..., fee_ratio: _Optional[_Union[FixedPointDecimal, _Mapping]] = ..., trade_id: _Optional[int] = ...) -> None: ...

class FixedPointDecimal(_message.Message):
    __slots__ = ("mantissa", "exponent")
    MANTISSA_FIELD_NUMBER: _ClassVar[int]
    EXPONENT_FIELD_NUMBER: _ClassVar[int]
    mantissa: int
    exponent: int
    def __init__(self, mantissa: _Optional[int] = ..., exponent: _Optional[int] = ...) -> None: ...

class AssetPosition(_message.Message):
    __slots__ = ("subaccount_id", "asset_id", "total", "available")
    SUBACCOUNT_ID_FIELD_NUMBER: _ClassVar[int]
    ASSET_ID_FIELD_NUMBER: _ClassVar[int]
    TOTAL_FIELD_NUMBER: _ClassVar[int]
    AVAILABLE_FIELD_NUMBER: _ClassVar[int]
    subaccount_id: int
    asset_id: int
    total: RawUnits
    available: RawUnits
    def __init__(self, subaccount_id: _Optional[int] = ..., asset_id: _Optional[int] = ..., total: _Optional[_Union[RawUnits, _Mapping]] = ..., available: _Optional[_Union[RawUnits, _Mapping]] = ...) -> None: ...

class RawUnits(_message.Message):
    __slots__ = ("word0", "word1", "word2", "word3")
    WORD0_FIELD_NUMBER: _ClassVar[int]
    WORD1_FIELD_NUMBER: _ClassVar[int]
    WORD2_FIELD_NUMBER: _ClassVar[int]
    WORD3_FIELD_NUMBER: _ClassVar[int]
    word0: int
    word1: int
    word2: int
    word3: int
    def __init__(self, word0: _Optional[int] = ..., word1: _Optional[int] = ..., word2: _Optional[int] = ..., word3: _Optional[int] = ...) -> None: ...

class Bootstrap(_message.Message):
    __slots__ = ("done", "resting", "position")
    DONE_FIELD_NUMBER: _ClassVar[int]
    RESTING_FIELD_NUMBER: _ClassVar[int]
    POSITION_FIELD_NUMBER: _ClassVar[int]
    done: Done
    resting: RestingOrders
    position: AssetPositions
    def __init__(self, done: _Optional[_Union[Done, _Mapping]] = ..., resting: _Optional[_Union[RestingOrders, _Mapping]] = ..., position: _Optional[_Union[AssetPositions, _Mapping]] = ...) -> None: ...

class RestingOrders(_message.Message):
    __slots__ = ("orders",)
    ORDERS_FIELD_NUMBER: _ClassVar[int]
    orders: _containers.RepeatedCompositeFieldContainer[RestingOrder]
    def __init__(self, orders: _Optional[_Iterable[_Union[RestingOrder, _Mapping]]] = ...) -> None: ...

class AssetPositions(_message.Message):
    __slots__ = ("positions",)
    POSITIONS_FIELD_NUMBER: _ClassVar[int]
    positions: _containers.RepeatedCompositeFieldContainer[AssetPosition]
    def __init__(self, positions: _Optional[_Iterable[_Union[AssetPosition, _Mapping]]] = ...) -> None: ...

class Done(_message.Message):
    __slots__ = ("latest_transact_time", "read_only")
    LATEST_TRANSACT_TIME_FIELD_NUMBER: _ClassVar[int]
    READ_ONLY_FIELD_NUMBER: _ClassVar[int]
    latest_transact_time: int
    read_only: bool
    def __init__(self, latest_transact_time: _Optional[int] = ..., read_only: bool = ...) -> None: ...

class RestingOrder(_message.Message):
    __slots__ = ("client_order_id", "exchange_order_id", "market_id", "price", "order_quantity", "side", "time_in_force", "order_type", "remaining_quantity", "rest_time", "subaccount_id", "cumulative_quantity", "cancel_on_disconnect")
    CLIENT_ORDER_ID_FIELD_NUMBER: _ClassVar[int]
    EXCHANGE_ORDER_ID_FIELD_NUMBER: _ClassVar[int]
    MARKET_ID_FIELD_NUMBER: _ClassVar[int]
    PRICE_FIELD_NUMBER: _ClassVar[int]
    ORDER_QUANTITY_FIELD_NUMBER: _ClassVar[int]
    SIDE_FIELD_NUMBER: _ClassVar[int]
    TIME_IN_FORCE_FIELD_NUMBER: _ClassVar[int]
    ORDER_TYPE_FIELD_NUMBER: _ClassVar[int]
    REMAINING_QUANTITY_FIELD_NUMBER: _ClassVar[int]
    REST_TIME_FIELD_NUMBER: _ClassVar[int]
    SUBACCOUNT_ID_FIELD_NUMBER: _ClassVar[int]
    CUMULATIVE_QUANTITY_FIELD_NUMBER: _ClassVar[int]
    CANCEL_ON_DISCONNECT_FIELD_NUMBER: _ClassVar[int]
    client_order_id: int
    exchange_order_id: int
    market_id: int
    price: int
    order_quantity: int
    side: Side
    time_in_force: TimeInForce
    order_type: OrderType
    remaining_quantity: int
    rest_time: int
    subaccount_id: int
    cumulative_quantity: int
    cancel_on_disconnect: bool
    def __init__(self, client_order_id: _Optional[int] = ..., exchange_order_id: _Optional[int] = ..., market_id: _Optional[int] = ..., price: _Optional[int] = ..., order_quantity: _Optional[int] = ..., side: _Optional[_Union[Side, str]] = ..., time_in_force: _Optional[_Union[TimeInForce, str]] = ..., order_type: _Optional[_Union[OrderType, str]] = ..., remaining_quantity: _Optional[int] = ..., rest_time: _Optional[int] = ..., subaccount_id: _Optional[int] = ..., cumulative_quantity: _Optional[int] = ..., cancel_on_disconnect: bool = ...) -> None: ...
